@@ -24,7 +24,7 @@ from itertools import combinations
 import matplotlib.pyplot as plt
 import json
 import ipywidgets as widgets
-
+import statsmodels.stats.multitest as smm
 
 class Network(object):
   '''
@@ -297,7 +297,7 @@ class Network(object):
 
     return export_data.export_net_json(self, which_viz, 'no-indent')
 
-  def widget(self, which_viz='viz', link_net=None, clust_library='scipy',
+  def widget(self, which_viz='viz', link_net=None, link_net_js=None, clust_library='scipy',
     min_samples=1, min_cluster_size=2):
     '''
     Generate a widget visualization using the widget. The export_viz_to_widget
@@ -334,7 +334,7 @@ class Network(object):
 
       self.widget_instance.observe(self.get_manual_category, names='manual_cat')
 
-    # add link
+    # add link (python)
     if link_net is not None:
       inst_link = widgets.link(
                                 (self.widget_instance, 'manual_cat'),
@@ -342,6 +342,15 @@ class Network(object):
                                )
       self.widget_instance.link = inst_link
 
+    # add jslink (JavaScript)
+    if link_net_js is not None:
+
+      inst_link = widgets.jslink(
+                                (self.widget_instance, 'manual_cat'),
+                                (link_net_js.widget_instance, 'manual_cat')
+                               )
+      self.widget_instance.link = inst_link
+
     return self.widget_instance
 
 
@@ -701,7 +710,7 @@ class Network(object):
     save_list_to_tsv(genes, inst_path + 'genes.tsv')
     save_list_to_tsv(barcodes, inst_path + 'barcodes.tsv')
 
-    mat_ge = df.get_values()
+    mat_ge = df.values
     mat_ge_sparse = sparse.coo_matrix(mat_ge)
 
     io.mmwrite( inst_path + 'matrix.mtx', mat_ge_sparse)
@@ -768,8 +777,8 @@ class Network(object):
           false_index = list(np.zeros(sim_dict['diff'].shape[0]))
           y_true = true_index + false_index
 
-          true_val = list(sim_dict['same'].get_values())
-          false_val = list(sim_dict['diff'].get_values())
+          true_val = list(sim_dict['same'].values)
+          false_val = list(sim_dict['diff'].values)
           y_score = true_val + false_val
 
           fpr, tpr, thresholds = roc_curve(y_true, y_score)
@@ -797,7 +806,105 @@ class Network(object):
 
       return sim_data
 
-  def generate_signatures(self, df_ini, category_level, pval_cutoff=0.05,
+  def generate_signatures(self, df_data, df_meta, category_name, pval_cutoff=0.05,
+                          num_top_dims=False, verbose=True, equal_var=False):
+
+      '''
+      Generate signatures for column categories
+      T-test is run for each category in a one-vs-all manner. The num_top_dims overrides
+      the P-value cutoff.
+      '''
+
+      df_t = df_data.transpose()
+
+      # remove columns (dimensions) with constant values
+      orig_num_cols = df_t.shape[1]
+      df_t = df_t.loc[:, (df_t != df_t.iloc[0]).any()]
+      if df_t.shape[1] < orig_num_cols:
+        print('dropped columns with constant values')
+
+      # make tuple rows
+      df_t.index = [(x, df_meta.loc[x, category_name]) for x in df_t.index.tolist()]
+      category_level = 1
+
+      df = self.row_tuple_to_multiindex(df_t)
+
+      cell_types = sorted(list(set(df.index.get_level_values(category_level).tolist())))
+
+      keep_genes = []
+      keep_genes_dict = {}
+      gene_pval_dict = {}
+      all_fold_info = {}
+
+      for inst_ct in cell_types:
+
+          inst_ct_mat = df.xs(key=inst_ct, level=category_level)
+          inst_other_mat = df.drop(inst_ct, level=category_level)
+
+          # save mean values and fold change
+          fold_info = {}
+          fold_info['cluster_mean'] = inst_ct_mat.mean()
+          fold_info['other_mean'] = inst_other_mat.mean()
+          fold_info['log2_fold'] = fold_info['cluster_mean']/fold_info['other_mean']
+          fold_info['log2_fold'] = fold_info['log2_fold'].apply(np.log2)
+          all_fold_info[inst_ct] = fold_info
+
+          inst_stats, inst_pvals = ttest_ind(inst_ct_mat, inst_other_mat, axis=0, equal_var=equal_var)
+
+          ser_pval = pd.Series(data=inst_pvals, index=df.columns.tolist()).sort_values()
+
+          if num_top_dims == False:
+              ser_pval_keep = ser_pval[ser_pval < pval_cutoff]
+          else:
+              ser_pval_keep = ser_pval[:num_top_dims]
+
+          gene_pval_dict[inst_ct] = ser_pval_keep
+
+          inst_keep = ser_pval_keep.index.tolist()
+          keep_genes.extend(inst_keep)
+          keep_genes_dict[inst_ct] = inst_keep
+
+      keep_genes = sorted(list(set(keep_genes)))
+
+      df_gbm = df.groupby(level=category_level).mean().transpose()
+      cols = df_gbm.columns.tolist()
+
+      df_sig = df_gbm.loc[keep_genes]
+
+      if len(keep_genes) == 0 and verbose:
+          print('found no informative dimensions')
+
+      df_gene_pval = pd.concat(gene_pval_dict, axis=1, sort=False)
+
+      df_diff = {}
+      for inst_col in df_gene_pval:
+
+          inst_pvals = df_gene_pval[inst_col]
+
+          # nans represent dimensions that did not meet pval or top threshold
+          inst_pvals = inst_pvals.dropna().sort_values()
+
+          inst_genes = inst_pvals.index.tolist()
+
+          # prevent failure if no cells have this category
+          if inst_pvals.shape[0] > 0:
+              rej, pval_corr = smm.multipletests(inst_pvals, 0.05, method='fdr_bh')[:2]
+
+              ser_pval = pd.Series(data=inst_pvals, index=inst_genes, name='P-values')
+              ser_bh_pval = pd.Series(data=pval_corr, index=inst_genes, name='BH P-values')
+              ser_log2_fold = all_fold_info[inst_col]['log2_fold'].loc[inst_genes]
+              ser_log2_fold.name = 'Log2 Fold Change'
+              ser_cluster_mean = all_fold_info[inst_col]['cluster_mean'].loc[inst_genes]
+              ser_cluster_mean.name = 'Cluster Mean'
+              ser_other_mean = all_fold_info[inst_col]['other_mean'].loc[inst_genes]
+              ser_other_mean.name = 'All Other Mean'
+              inst_df = pd.concat([ser_pval, ser_bh_pval, ser_log2_fold, ser_cluster_mean, ser_other_mean], axis=1)
+
+              df_diff[inst_col] = inst_df
+
+      return df_sig, df_diff
+
+  def old_generate_signatures(self, df_ini, category_level, pval_cutoff=0.05,
                           num_top_dims=False, verbose=True, equal_var=False):
 
       ''' Generate signatures for column categories '''
@@ -864,9 +971,48 @@ class Network(object):
 
       df_gene_pval = pd.concat(gene_pval_dict, axis=1, sort=False)
 
-      return df_sig, keep_genes_dict, df_gene_pval, all_fold_info
+      return df_sig, df_gene_pval, all_fold_info
 
-  def predict_cats_from_sigs(self, df_data_ini, df_sig_ini, dist_type='cosine', predict_level='Predict Category',
+  def predict_cats_from_sigs(self, df_data_ini, df_meta, df_sig_ini,
+                             predict='Predicted Category', dist_type='cosine',
+                             unknown_thresh=-1):
+      ''' Predict category using signature '''
+
+      keep_rows = df_sig_ini.index.tolist()
+      data_rows = df_data_ini.index.tolist()
+
+      common_rows = list(set(data_rows).intersection(keep_rows))
+
+      df_data = deepcopy(df_data_ini.loc[common_rows])
+      df_sig = deepcopy(df_sig_ini.loc[common_rows])
+
+      # calculate sim_mat of df_data and df_sig
+      cell_types = df_sig.columns.tolist()
+      barcodes = df_data.columns.tolist()
+      sim_mat = 1 - pairwise_distances(df_sig.transpose(), df_data.transpose(), metric=dist_type)
+      df_sim = pd.DataFrame(data=sim_mat, index=cell_types, columns=barcodes).transpose()
+
+      # get the top column value (most similar signature)
+      df_sim_top = df_sim.idxmax(axis=1)
+
+      # get the maximum similarity of a cell to a cell type definition
+      max_sim = df_sim.max(axis=1)
+
+      unknown_cells = max_sim[max_sim < unknown_thresh].index.tolist()
+
+      # assign unknown cells (need category of same name)
+      df_sim_top[unknown_cells] = 'Unknown'
+
+      # add predicted category name to top list
+      # top_list = df_sim_top.values
+      df_sim = df_sim.transpose()
+
+      df_meta[predict] = df_sim_top
+
+      return df_sim
+
+
+  def old_predict_cats_from_sigs(self, df_data_ini, df_sig_ini, dist_type='cosine', predict_level='Predict Category',
                              truth_level=1, unknown_thresh=-1):
       ''' Predict category using signature '''
 
@@ -896,7 +1042,7 @@ class Network(object):
       df_sim_top[unknown_cells] = 'Unknown'
 
       # add predicted category name to top list
-      top_list = df_sim_top.get_values()
+      top_list = df_sim_top.values
       top_list = [ predict_level + ': ' + x[0] if type(x) is tuple else predict_level + ': ' + x  for x in top_list]
 
       # add cell type category to input data
@@ -928,7 +1074,36 @@ class Network(object):
 
       return df_cat, df_sim.transpose(), y_info
 
-  def confusion_matrix_and_correct_series(self, y_info):
+  def assess_prediction(self, df_meta, truth, pred):
+      ''' Generate confusion matrix from y_info '''
+
+      y_info = {}
+      y_info['true'] = df_meta[truth].values.tolist()
+      y_info['pred'] = df_meta[pred].values.tolist()
+
+      sorted_cats = sorted(list(set(y_info['true'] + y_info['pred'])))
+      conf_mat = confusion_matrix(y_info['true'], y_info['pred'], labels=sorted_cats)
+
+      # true columns and pred rows
+      df_conf = pd.DataFrame(conf_mat, index=sorted_cats, columns=sorted_cats).transpose()
+
+      total_correct = np.trace(df_conf)
+      total_pred = df_conf.sum().sum()
+      fraction_correct = total_correct/float(total_pred)
+
+      # calculate ser_correct
+      correct_list = []
+      cat_counts = df_conf.sum(axis=0)
+      all_cols = df_conf.columns.tolist()
+      for inst_cat in all_cols:
+          inst_correct = df_conf.loc[inst_cat, inst_cat] / cat_counts[inst_cat]
+          correct_list.append(inst_correct)
+
+      ser_correct = pd.Series(data=correct_list, index=all_cols)
+
+      return df_conf, ser_correct, fraction_correct
+
+  def old_confusion_matrix_and_correct_series(self, y_info):
       ''' Generate confusion matrix from y_info '''
 
 
@@ -980,7 +1155,7 @@ class Network(object):
 
           cols = df_data.columns.tolist()
           rows = df_data.index.tolist()
-          mat = df_data.get_values()
+          mat = df_data.values
 
           shuffled_cols = deepcopy(cols)
           random.shuffle(shuffled_cols)
